/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 add_noise.c tiffread.c tiffwrite.c -lm -ltiff -o add_noise

  References :

  [1] R. Abergel and L. Moisan, ``The Shannon Total Variation'', preprint MAP5 n° 2016-19, July 2016.
  [2] L. Moisan. How to discretize the total variation of an image? Proceedings in Applied Mathematics and Mechanics, 2007.

  Versions :

  + v 1.0 : (09/2016) initial version, adapted from MegaWave (Rémy Abergel)

 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#define VERSION "1.0"

#define FREE_ALL() {				\
    if(NULL != u) free(u);			\
    if(NULL != out) free(out);			\
  }
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }

static void display_usage()
{
  printf("\nAdd white Gaussian noise with zero mean to an image.\n\n");
  printf("Usage: add_noise [-g std] [-f] in out\n\n");
  printf("   -g std  : (default 10.) standard deviation of the Gaussian noise\n");
  printf("   -f      : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   in      : input image (TIFF)\n");
  printf("   out     : output image (TIFF)\n\n");
}


int main(int argc, char **argv)
{
  char *filename_in = NULL, *filename_out = NULL, *gvalue = NULL, vflag, fflag;
  double *u = NULL,*out = NULL,std,a,b;
  int i,nx,ny,err,adr;
  vflag = fflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (strcmp(argv[i],"-g") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -g requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	gvalue = argv[i+1]; err = sscanf(gvalue,"%lf",&std); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -g\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (NULL == filename_in) filename_in = argv[i];
    else if (NULL == filename_out) filename_out = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* check consistency */
  if ((NULL == filename_in) ||  (NULL == filename_out)) {
    display_usage();
    if (NULL == filename_in) printf("Error: input 'in' is missing\n\n");
    else printf("Error: output 'out' is missing\n\n");
    return EXIT_FAILURE;
  }

  /* set default value(s) */
  if (NULL == gvalue) std = 10.;

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(EXIT_FAILURE == read_tiff_into_double(&u,&nx,&ny,filename_in,vflag)) {
    printf("An error occured during the TIFF image reading process.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  ASSERT_ALLOC(out = (double*) malloc (nx*ny * sizeof(double)));



  /***********************************/
  /*              KERNEL             */
  /***********************************/
  srand48( (long int) time (NULL) + (long int) getpid() );
  for (adr=0;adr<nx*ny;adr++) {
      a = drand48();
      b = drand48();
      out[adr] = u[adr] + std*sqrt(-2.0*log(a))*cos(2.0*M_PI*b);
    }
  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/


  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if (EXIT_FAILURE == save_tiff(filename_out,out,nx,ny,8,fflag,1,vflag)) {
    printf("An error occured during the TIFF image saving process.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }


  /* free memory */
  FREE_ALL();
}
