/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 tiffextract.c tiffread.c tiffwrite.c -lm -ltiff -o tiffextract

  References :

  [1] R. Abergel and L. Moisan, ``The Shannon Total Variation'', preprint MAP5 n° 2016-19, July 2016.
  [2] L. Moisan. How to discretize the total variation of an image? Proceedings in Applied Mathematics and Mechanics, 2007.

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel), adapted from MegaWave module fextract.

 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#define VERSION "1.0"

#define FREE_ALL() {				\
    if(NULL != u) free(u);			\
    if(NULL != out) free(out);			\
  }

#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }
#define MIN(a,b)        (((a)>(b)) ? (b) : (a))

static void display_usage()
{
  printf("\nExtract a subpart of a TIFF image.\n\n");
  printf("Usage: tiffextract [-r] [-f] in out X1 Y1 X2 Y2\n\n");
  printf("   -r    : if set, X2 and Y2 must be the SIZE of the extracted region\n");
  printf("   -f    : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   in    : input image (TIFF)\n");
  printf("   out   : output image (TIFF)\n");
  printf("   X1    : upleft corner of the region to extract from input (x)\n");
  printf("   Y1    : upleft corner of the region to extract from input (y)\n");
  printf("   X2    : downright corner of the region to extract from input (x)\n");
  printf("   Y2    : downright corner of the region to extract from input (y)\n\n");
}


int main(int argc, char **argv)
{
  char *filename_in = NULL, *filename_out = NULL, *X1value = NULL, *X2value = NULL, *Y1value = NULL, *Y2value = NULL, rflag, vflag, fflag;
  double *u = NULL,*out = NULL;
  int i,nx,ny,X1,X2,Y1,Y2,x,y,adr,err,nx_out,ny_out;
  rflag = vflag = fflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (strcmp(argv[i],"-r") == 0) rflag = 1;
    else if (NULL == filename_in) filename_in = argv[i];
    else if (NULL == filename_out) filename_out = argv[i];
    else if (NULL == X1value) {
      X1value = argv[i]; err = sscanf(X1value,"%d",&X1);
      if(err != 1) { display_usage(); printf("Error: could not retrieve properly input 'X1'\n\n"); return EXIT_FAILURE; }
    }
    else if (NULL == Y1value) {
      Y1value = argv[i]; err = sscanf(Y1value,"%d",&Y1);
      if(err != 1) { display_usage(); printf("Error: could not retrieve properly input 'Y1'\n\n"); return EXIT_FAILURE; }
    }
    else if (NULL == X2value) {
      X2value = argv[i]; err = sscanf(X2value,"%d",&X2);
      if(err != 1) { display_usage(); printf("Error: could not retrieve properly input 'X2'\n\n"); return EXIT_FAILURE; }
    }
    else if (NULL == Y2value) {
      Y2value = argv[i]; err = sscanf(Y2value,"%d",&Y2);
      if(err != 1) { display_usage(); printf("Error: could not retrieve properly input 'Y2'\n\n"); return EXIT_FAILURE; }
    }
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* check consistency */
  if ((NULL == filename_in) || (NULL == filename_out) || (NULL == X1value) || (NULL == X2value) || (NULL == Y1value) || (NULL == Y2value)) {
    display_usage();
    if (NULL == filename_in) printf("Error: input 'in' is missing\n\n");
    else if (NULL == filename_out) printf("Error: output 'out' is missing\n\n");
    else if (NULL == X1value) printf("Error: input 'X1' is missing\n\n");
    else if (NULL == X2value) printf("Error: input 'X2' is missing\n\n");
    else if (NULL == Y1value) printf("Error: input 'Y1' is missing\n\n");
    else printf("Error: input 'Y2' is missing\n\n");
    return EXIT_FAILURE;
  }

  if ((!rflag)&&(X2<=X1)) { printf("\nError: X2 <= X1 is not allowed (except using option -r).\n\n"); return EXIT_FAILURE; }
  if ((!rflag)&&(Y2<=Y1)) { printf("\nError: X2 <= X1 is not allowed (except using option -r).\n\n"); return EXIT_FAILURE; }
  if (X1<0) { printf("\nError: input 'X1' must be nonnegative.\n\n"); return EXIT_FAILURE; }
  if (X2<0) { printf("\nError: input 'X2' must be nonnegative.\n\n"); return EXIT_FAILURE; }
  if (Y1<0) { printf("\nError: input 'Y1' must be nonnegative.\n\n"); return EXIT_FAILURE; }
  if (Y2<0) { printf("\nError: input 'Y2' must be nonnegative.\n\n"); return EXIT_FAILURE; }

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(EXIT_FAILURE == read_tiff_into_double(&u,&nx,&ny,filename_in,vflag)) {
    printf("An error occured during the TIFF image reading process.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  nx_out = (rflag) ? X2 : X2-X1+1;
  ny_out = (rflag) ? Y2 : Y2-Y1+1;
  ASSERT_ALLOC(out = (double*) calloc (nx_out*ny_out , sizeof(double)));



  /***********************************/
  /*             KERNEL              */
  /***********************************/
  for(x=X1;x<=MIN(((rflag) ? X1+nx_out-1 : X2),nx-1);x++)
    for(y=Y1;y<=MIN(((rflag) ? Y1+ny_out-1 : Y2),nx-1);y++)
      out[(y-Y1)*nx_out + (x-X1)] = u[y*nx+x];
  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/


  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if (EXIT_FAILURE == save_tiff(filename_out,out,nx_out,ny_out,8,fflag,1,vflag)) {
    printf("An error occured during the TIFF image saving process.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }


  /* free memory */
  FREE_ALL();
}
